;*DDK*************************************************************************/
;
; COPYRIGHT (C) Microsoft Corporation, 1989
; COPYRIGHT    Copyright (C) 1995 IBM Corporation
;
;    The following IBM OS/2 WARP source code is provided to you solely for
;    the purpose of assisting you in your development of OS/2 WARP device
;    drivers. You may use this code in accordance with the IBM License
;    Agreement provided in the IBM Device Driver Source Kit for OS/2. This
;    Copyright statement may not be removed.;
;*****************************************************************************/
;/*****************************************************************************
;*
;* SOURCE FILE NAME = SCANLINE.INC
;*
;* DESCRIPTIVE NAME = definitions for scanline
;*
;*
;* VERSION      V2.0
;*
;* DATE         8/4/92
;*
;* DESCRIPTION  
;*
;* FUNCTIONS    NONE
;*              
;*
;* NOTES        NONE
;*
;* STRUCTURES   NONE
;*
;* EXTERNAL REFERENCES
;*
;*              NONE
;*
;* EXTERNAL FUNCTIONS
;*
;*              NONE
;*
;* CHANGE ACTIVITY =
;*   DATE      FLAG        APAR   CHANGE DESCRIPTION
;*   --------  ----------  -----  --------------------------------------
;*   mm/dd/yy  @Vr.mpppxx  xxxxx  xxxxxxx
;*
;*****************************************************************************/



NBR_RECTS       equ     4       ; # of clip-Rects in buffer on stack

EXTENDED_CONTROL struc
        ec_control      dw      (size RGNRECT)/2 dup (?)
        ec_buffer       dw      NBR_RECTS * (size RECTL ) / 2 dup (?)
        ec_rect_bound   dw      (size RECTL)/2 dup (?)
EXTENDED_CONTROL ends


;/***************************************************************************
;*
;* MACRO NAME    = imin
;*
;* DESCRIPTION   = Returns in R1 the smallest of R1 and R2.  Note that these   
;*                 are treated as signed integers.
;*
;*                 Registers Destroyed:
;*                       Flags
;*                 Registers Preserved:
;*                       All
;*
;* INPUT         = NONE
;* OUTPUT        = R1 = min(R1,R2) 
;*
;* RETURN-NORMAL = NONE
;* RETURN-ERROR  = NONE
;*
;**************************************************************************/

imin    macro   R1,R2
        local   got_min_already
        cmp     R1,R2
        jl      got_min_already
        mov     R1,R2
got_min_already:
        endm

;/***************************************************************************
;*
;* MACRO NAME    = imax
;*
;* DESCRIPTION   = Returns in R1 the largest of R1 and R2.  Note that these
;*                 are treated as signed integers.
;*
;*                 Registers Destroyed:
;*                       Flags
;*                 Registers Preserved:
;*                       All
;*
;* INPUT         = NONE
;* OUTPUT        = R1 = max(R1,R2)
;*
;* RETURN-NORMAL = NONE
;* RETURN-ERROR  = NONE
;*
;**************************************************************************/


imax    macro   R1,R2
        local   got_max_already
        cmp     R1,R2
        jg      got_max_already
        mov     R1,R2
got_max_already:
        endm



;/*
;**     The cntrl_blk structure will contain the addresses of the
;**     processor for the first and last byte, inner loop, and the
;**     accelerator byte for skipping the operation.
;**
;**     There will be one structure for each plane processed
;*/

cntrl_blk       struc
first_last_proc dw      ?               ;Address for first and last byte
inner_loop_proc dw      ?               ;Address for inner loop bytes
nop_flag        db      ?               ;D7 set if this is a nop
pattern         db      ?               ;The pattern for the plane
cntrl_blk       ends

                .errnz  (size cntrl_blk) and 1

;/*
;**     Bit definitions within the nop_flag field in the cntrl_blk
;**     structure.
;*/

NF_NOP          equ     00000001b
;**             equ     00000010b
;**             equ     00000100b
;**             equ     00001000b
;**             equ     00010000b
;**             equ     00100000b
;**             equ     01000000b
;**             equ     10000000b


;/*
;**     The following values are used to index into the control
;**     block structures for each plane.  The value for C3_DATA
;**     will be incorrect for a three plane driver, but then it
;**     won't be used for a three plane driver.
;**
;**     The values are stored in reverse order to match the
;**     looping logic with the plane selection logic.
;*/


;/*
;**     The following equates are for the binary raster ops
;**     which will be passed to this routine.  The values
;**     shown are 0:15 since the rop is mapped upon entry
;**     into this function.
;*/


ROP_DDX         equ     0               ;DDx
ROP_DPON        equ     1               ;DPon
ROP_DPNA        equ     2               ;DPna
ROP_PN          equ     3               ;Pn
ROP_PDNA        equ     4               ;PDna
ROP_DN          equ     5               ;Dn
ROP_DPX         equ     6               ;DPx
ROP_DPAN        equ     7               ;DPan
ROP_DPA         equ     8               ;DPa
ROP_DPXN        equ     9               ;DPxn
ROP_D           equ     10              ;D
ROP_DPNO        equ     11              ;DPno
ROP_P           equ     12              ;P
ROP_PDNO        equ     13              ;PDno
ROP_DPO         equ     14              ;DPo
ROP_DDXN        equ     15              ;DDxn


;/*
;**     The following values will be set in another_tbl for
;**     doing various accelerations on the raster op.
;**
;**     NO_OBJECT is the value used to indicate that the drawing
;**     operation doesn't require a pen or brush.
;**
;**     STOSB_OK is the value used to show that the inner loop
;**     code for the EGA can use STOSB for the given rop as long
;**     as there are no transparent pixels.
;**
;**     MY_SINGLE_OK is set to indicate that the raster operation
;**     may be performed in one pass on the EGA if the brush is
;**     solid.  This is identical to SINGLE_OK, by why do extra
;**     table lookups?
;**
;**     NEG_PATTERN is used as a flag for special cased P and Pn
;**     operations on the EGA.  If this flag is set, then the
;**     pattern needs to be negated before stored in EGA memory
;**     at current_brush.
;**
;**     ALT_NEG_PATTERN is used as a flag to indicated that the
;**     pattern should be negated for word templates.
;*/


NEG_PATTERN     equ     10000000b
NO_OBJECT       equ     01000000b
MY_SINGLE_OK    equ     00100000b
STOSB_OK        equ     00010000b
ALT_NEG_PATTERN equ     00000001b



;/*
;**     Define the loc_clean_up flags used to indicate what actions
;**     are required upon exit for restoring the EGA to the
;**     default state.
;**
;**     The flag will also be overloaded to contain a flag to
;**     the outer loop code indicating that EGA plane selection
;**     is required.
;*/


CU_NONE         equ     00000000b       ;No cleanup is needed
CU_REGS         equ     10000000b       ;Registers need restoring
CU_EXCLUDE      equ     01000000b       ;Exclusion rectangle must be cleared
CU_PLANE_SEL    equ     00000001b       ;EGA plane selection needed


;/*
;**     The following are the values which will be stored
;**     in loc_some_flags.
;*/

INDEX_MONO      equ     10000000b       ;Device is monochrome
INDEX_BLNK_FORGRND equ  00000100b       ;Need no paint on foreground
INDEX_BLNK_BKGRND  equ  00000010b       ;Need no paint on background
INDEX_BKGRD     equ     00000001b       ;Processors for background mixmode


MF_LAST_PASS     equ     00000001b
MF_KILL_SL       equ     00000010b
MF_OPQ_CLP_LFT   equ     00000100b
MF_OPQ_CLP_RGT   equ     00001000b
MF_FGNOP_BKOPQ   equ     00010000b       ;if fg mix is NOP, bk mix is OPAQUE
MF_UNEXCLUDE     equ     00100000b
MF_USE_DDC_RECTS equ     01000000b
;**              equ     10000000b

;/*
;**     The following are the values which will be stored in
;**     loc_brush_accel during the color bitmap polyscanlines
;**     execution.
;*/

FG_IS_LEAVEALONE        equ     00000001b       ; fg mix is LeaveAlone
FG_MIX_IS_NOT_BG_MIX    equ     00000010b       ; fg mix differs from bg mix
;**                     equ     00000100b
FG_COLOR_IS_NOT_BG_COLOR equ    00001000b       ; fg,bg colors not the same
BRUSH_IS_PATTERNED      equ     00010000b       ; brush is not solid
BG_IS_LEAVEALONE        equ     00100000b       ; bg mix is LeaveAlone
FG_IS_NOT_REPLACE       equ     01000000b       ; fg mix is not replace
;**                     equ     10000000b
REP_STOS_IS_OK          equ     00011010b       ; if all 0s then ok to rep stos

;/***************************************************************************
;*
;* MACRO NAME    = define_frame
;*
;* DESCRIPTION   = Macro for defining the frame for PolyScanline functions   
;*                                                                           
;*                 This file allows the frame definition to be shared by all 
;*                 the modules which make up the PolyScanline function.      
;*
;* INPUT         = NONE
;* OUTPUT        = NONE
;*
;* RETURN-NORMAL = NONE
;* RETURN-ERROR  = NONE
;*
;**************************************************************************/

define_frame macro  name,nearness,forcefar

ifb     <forcefar>
cProc   name,<nearness,PUBLIC,WIN,PASCAL,NODATA>,<si,di>
else
if      forcefar
cProc   name,<FAR,PUBLIC,WIN,PASCAL,NODATA>,<si,di>
else
cProc   name,<NEAR,PUBLIC,WIN,PASCAL,NODATA>,<si,di>
endif
endif

        parmD   hdc
        parmD   pScansData
        parmD   hddc
        parmD   FunN

        localD  arg_lp_ddc              ;--> device display context

        localD  loc_right_sl
        localD  loc_left_sl
        localD  loc_right_sl_ini
        localD  loc_left_sl_ini
        localD  loc_right_sl_x
        localD  loc_left_sl_x
        localD  loc_points
        localW  loc_sl_count
        localW  loc_sl_count_ini
        localW  loc_scan_count          ; # of deltas in PSL
        localW  loc_saved_y             ; Scan's Y coordinate (our perspective)
        localW  loc_saved_y_unfliped    ; Scan's Y coord   (Engine perspective)
        localW  loc_next_plane          ; Index to next plane of the scan
        localW  loc_raster_op           ; Rop2, mapped 0==>15
        localW  loc_bk_raster_op        ; Rop of BackGround
        localW  loc_bm_height
        localW  loc_current_rect
        localD  loc_scan_start
        localV  loc_control,%(size EXTENDED_CONTROL)
        localV  loc_processors,%(size cntrl_blk)
        localV  loc_bk_processors,%(size cntrl_blk)
        localW  loc_proc_offset         ; Offset to the first control block elem
        localW  loc_transparency_mask            ; Mask of altered bits within a byte
        localW  loc_bk_transparency_mask; Mask of altered bits within a byte
        localB  loc_some_flags          ; Some flags
        localB  loc_moore_flags         ; Moore flags
        localB  loc_clean_up            ; Flags for cleaning up at exit
        localW  loc_first_mask
        localW  loc_last_mask
        localW  loc_inner_loop_count
        localW  rc                      ;return code

        localB  loc_brush_accel         ; local brush flags
        localW  loc_pat_row             ; pattern origin in y
        localW  loc_PatCacheY           ; top y of board-cached pattern
        localD  loc_scanline_init       ; -> actual init'er at bottom level
        localD  loc_scanline_doer       ; -> actual do'er at bottom level
        localD  loc_scanline_deinit     ; -> actual deinit'er at bottom level
        localW  loc_both_colors         ; bg,fg colors in case solid pattern
        localW  loc_dev_cmd             ; cmd to send to hw
        localV  loc_trans_mask,SIZE_PATTERN; local bucket for brush mask
        localD  loc_compiled_proc       ; far -> compiled code procedure

        endm
